% Example10_1.m
% SNR scalable video coding using motion compensated prediction
% Block motion is estimated to an integer pel accuracy
% using full search and SAD matching metric.
% Block size is 8 x 8 pixels.
% Search window size is 16 x 16 pixels.
% The differential block is 2D DCT transformed, quantized
% using uniform quantizer with constant quantization step of 16.
% The reconstructed block becomes the reference block for
% the next input frame.
% Base layer carries the above quantized data.
% The enhanced layer is created by coding the difference
% between the unquantized DCT coefficients and the base layer
% quantized DCT coefficients and quantizing this differential
% DCT qith a different quantization step.
% Decoding the base layer yields a lower quality video and
% adding the enhanced layer to the base layer results in a
% higher quality video.
% Only intensity (Luma) image sequence is accepted.
% Note: Both Table Tennis and Trevor sequences have ".ras" 
% extension while the Claire sequence has ".yuv" extension.


clear
N = 8;% block size is N x N pixels
N2 = 2*N;
W = 16; % search window size is W x W pixels
quantizer_scale = 4; % used only for the base layer
% Four different sequences are tested.
% "rhinos.avi" exists in MATLAB and the others do not.
%inFile = 'tt040.ras';% Table Tennis sequence
%inFile = 'twy040.ras'; % Trevor sequence
%inFile = 'clairey040.yuv'; % Claire sequence
inFile = 'rhinos.avi'; % AVI file from MATLAB
strIndx = regexp(inFile,'\.');
if strcmpi('avi',inFile(strIndx+1:strIndx+3))
    M = aviread(inFile); % There are 114 320x240x3 frames
    F = int16(1:10);% frames to encode
else
    F = int16(40:49);% frames to encode
    M.cdata = cell(1,length(F));
    for k = 1:length(F)
        strIndx1 = regexp(inFile,'0');
        inFile1 = strcat(inFile(1:strIndx1(1)),num2str(F(k)),...
            inFile(strIndx:end));
        M(k).cdata = imread(inFile1);
    end
end
% Make image size divisible by 8
[X,Y,Z] = size(M(1).cdata);
if mod(X,8) ~=0
    Height = floor(X/8)*8;
else
    Height = X;
end
if mod(Y,8) ~=0
    Width = floor(Y/8)*8;
else
    Width = Y;
end
Depth = Z;
clear X Y Z
%
if Depth == 3
    A = rgb2ycbcr(M(1).cdata);% Convert RGB to YCbCr & retain only Y
    y_ref = A(:,:,1);
else
    A = M(1).cdata;
    y_ref = A;
end
% pad the reference frame left & right and top & bottom
y_ref = double(padarray(y_ref,[W/2 W/2],'replicate'));
% arrays to store SNR and PSNR values
Base_snr = zeros(1,length(F)-1); Enhanced_snr = zeros(1,length(F)-1);
Base_psnr = zeros(1,length(F)-1); Enhanced_psnr = zeros(1,length(F)-1);
% Encode the monochrome video using MPC
for f = 2:length(F)
    if Depth == 3
        B = rgb2ycbcr(M(f).cdata);
        y_current = B(:,:,1);
    else
        y_current = M(f).cdata;
    end
    y_current = double(padarray(y_current,[W/2 W/2],'replicate'));
    for r = N:N:Height
        rblk = floor(r/N);
        for c = N:N:Width
            cblk = floor(c/N);
            D = 1.0e+10;% initial city block distance
            for u = -N:N
                for v = -N:N
                    d = y_current(r+1:r+N,c+1:c+N)-y_ref(r+u+1:r+u+N,c+v+1:c+v+N);
                    d = sum(abs(d(:)));% city block distance between pixels
                    if d < D
                        D = d;
                        x1 = v; y1 = u; % motion vector
                    end
                end
            end
            % MC compensated difference coding
            temp = y_current(r+1:r+N,c+1:c+N)...
                -y_ref(r+1+y1:r+y1+N,c+1+x1:c+x1+N);
            TemP = dct2(temp); % DCT of difference
            s = sign(TemP); % extract the coefficient sign
            TemP1 = s .* round(abs(TemP)/(16*quantizer_scale))...
                *(16*quantizer_scale); % quantize/dequantize DCT
            temp = idct2(TemP1); % IDCT
            Base(r-N+1:r,c-N+1:c) = y_ref(r+1+y1:r+y1+N,c+1+x1:c+x1+N) +...
                temp; % reconstructed block - base quality
            delta_DCT = TemP - TemP1; % incremental DCT
            s1 = sign(delta_DCT); % extract the sign of incremental DCT
            delta_DCT = s1 .* round(abs(delta_DCT)/...
                4)*4;
            temp1 = idct2(TemP1 + delta_DCT);
            Enhanced(r-N+1:r,c-N+1:c) = y_ref(r+1+y1:r+y1+N,c+1+x1:c+x1+N) +...
                temp1;
        end
    end
    % Calculate the respective SNRs and PSNRs
    Base_snr(f-1) = 20*log10(std2(y_current(N+1:Height+N,N+1:Width+N))/...
        std2(y_current(N+1:Height+N,N+1:Width+N)-Base));
    Enhanced_snr(f-1) = 20*log10(std2(y_current(N+1:Height+N,N+1:Width+N))/...
        std2(y_current(N+1:Height+N,N+1:Width+N)-Enhanced));
    Base_psnr(f-1) = 20*log10(255/std2(y_current(N+1:Height+N,N+1:Width+N)-Base));
    Enhanced_psnr(f-1) = 20*log10(255/std2(y_current(N+1:Height+N,N+1:Width+N)...
        -Enhanced));
    % replace previous frames by the currently reconstructed frames
    y_ref = Base;
    y_ref = double(padarray(y_ref,[W/2 W/2],'replicate'));
end
figure,plot(F(2:end),Base_snr,'k*','LineWidth',1), hold on
plot(F(2:end),Enhanced_snr,'kd','LineWidth',2), title('SNR (dB)')
axis([F(2) F(end) min(Base_snr)-2 max(Enhanced_snr)+2]) % for Rhinos sequence
legend('Base Quality','Enhanced Quality',0)
xlabel('Frame #'), ylabel('SNR (dB)'), hold off
figure,plot(F(2:end),Base_psnr,'k*','LineWidth',1), hold on
plot(F(2:end),Enhanced_psnr,'kd','LineWidth',2), title('PSNR (dB)')
axis([F(2) F(end) min(Base_psnr)-2 max(Enhanced_psnr)+2]) % for Rhinos sequence
legend('Base Quality','Enhanced Quality',0)
xlabel('Frame #'), ylabel('PSNR (dB)'), hold off