% Example10_2.m
% Spatially scalable video coding using motion compensated prediction
% Block motion is estimated to an integer pel accuracy
% using full search and SAD matching metric.
% Block size is 8 x 8 pixels.
% Search window size is 16 x 16 pixels.
% The differential block is 2D DCT transformed, quantized
% using uniform quantizer with constant quantization step of 16.
% The reconstructed block becomes the reference block for
% the next input frame.
% Base layer carries the above quantized data.
% The enhanced layer is created by upsampling the base layer 
% reference image and subtracting it from the current full 
% resolution picture, taking the block DCT of the difference image, 
% quantizing and VLC coding.
% Decoding the base layer yields a lower spatial resolution video and
% adding the enhanced layer to the upsampled base layer results in a
% higher resolution video.
% Only intensity (Luma) image sequence is accepted.
% Note: Both Table Tennis and Trevor sequences have ".ras" 
% extension while the Claire sequence has ".yuv" extension.


clear
N = 8;% block size is N x N pixels
N2 = 2*N;
W = 16; % search window size is W x W pixels
quantizer_scale = 2; % used for both the base and enhanced layers
Interp_type = 'bicubic'; % type of interpolation used for upsampling
% Four different sequences are tested.
% "rhinos.avi" exists in MATLAB and the others do not.
%inFile = 'tt040.ras';% Table Tennis sequence
%inFile = 'twy040.ras'; % Trevor sequence
%inFile = 'clairey040.yuv'; % Claire sequence
inFile = 'rhinos.avi'; % AVI file from MATLAB
strIndx = regexp(inFile,'\.');
if strcmpi('avi',inFile(strIndx+1:strIndx+3))
    M = aviread(inFile); % There are 114 320x240x3 frames
    F = int16(1:10);% frames to encode
else
    F = int16(40:49);% frames to encode
    M.cdata = cell(1,length(F));
    for k = 1:length(F)
        strIndx1 = regexp(inFile,'0');
        inFile1 = strcat(inFile(1:strIndx1(1)),num2str(F(k)),...
            inFile(strIndx:end));
        M(k).cdata = imread(inFile1);
    end
end
% Make image size divisible by 8
[X,Y,Z] = size(M(1).cdata);
if mod(X,8) ~=0
    Height = floor(X/8)*8;
else
    Height = X;
end
if mod(Y,8) ~=0
    Width = floor(Y/8)*8;
else
    Width = Y;
end
Depth = Z;
clear X Y Z
base_Height = 136; % height of lower resolution image
base_Width = 184;  % width of lower resolution image
%
if Depth == 3
    A = rgb2ycbcr(M(1).cdata);% Convert RGB to YCbCr & retain only Y
    y_ref_enhance = double(A(:,:,1));
    y_ref_base = imresize(A(:,:,1),[base_Height base_Width],Interp_type);
else
    A = M(1).cdata;
    y_ref_enhance = double(A);
    y_ref_base = imresize(A,[base_Height base_Width],Interp_type);
end
% pad the reference frame left & right and top & bottom
y_ref_base = double(padarray(y_ref_base,[W/2 W/2],'replicate'));
% arrays to store SNR and PSNR values
Base_snr = zeros(1,length(F)-1); Enhanced_snr = zeros(1,length(F)-1);
Base_psnr = zeros(1,length(F)-1); Enhanced_psnr = zeros(1,length(F)-1);
% Encode the monochrome video using MPC
for f = 2:length(F)
    if Depth == 3
        B = rgb2ycbcr(M(f).cdata);
        y_current_enhance = double(B(:,:,1));
        y_current_base = imresize(B(:,:,1),...
            [base_Height base_Width],Interp_type);
    else
        y_current_enhance = double(M(f).cdata);
        y_current_base = imresize(M(f).cdata,...
            [base_Height base_Width],Interp_type);
    end
    y_current_base = double(padarray(...
        y_current_base,[W/2 W/2],'replicate'));
    for r = N:N:base_Height
        for c = N:N:base_Width
            D = 1.0e+10;% initial city block distance
            for u = -N:N
                for v = -N:N
                    d = y_current_base(r+1:r+N,c+1:c+N)-...
                        y_ref_base(r+u+1:r+u+N,c+v+1:c+v+N);
                    d = sum(abs(d(:)));% city block distance between pixels
                    if d < D
                        D = d;
                        x1 = v; y1 = u; % motion vector
                    end
                end
            end
            % MC compensated difference coding
            temp = y_current_base(r+1:r+N,c+1:c+N)...
                -y_ref_base(r+1+y1:r+y1+N,c+1+x1:c+x1+N);
            TemP = dct2(temp); % DCT of difference
            s = sign(TemP); % extract the coefficient sign
            TemP = s .* round(abs(TemP)/(16*quantizer_scale))...
                *(16*quantizer_scale); % quantize/dequantize DCT
            temp = idct2(TemP); % IDCT
            Base(r-N+1:r,c-N+1:c) =...
                y_ref_base(r+1+y1:r+y1+N,c+1+x1:c+x1+N) +...
                temp; % reconstructed block - base quality
        end
    end
    % Generate enhancement
    Delta = y_current_enhance - imresize(Base,[Height Width],Interp_type);
    for r = 1:N:Height
        for c = 1:N:Width
            temp = Delta(r:r+N-1,c:c+N-1);
            temp_DCT = dct2(temp);
            s = sign(temp_DCT);
            temp_DCT = s .* round(abs(temp_DCT)/(16*quantizer_scale))...
                * (16*quantizer_scale);
            E(r:r+N-1,c:c+N-1) = idct2(temp_DCT);
        end
    end
    Enhanced = E + imresize(Base,[Height Width],Interp_type);
    % replace previous frame by the currently reconstructed frame
    y_ref_base = Base;
    % pad the reference frame left & right and top & bottom
    y_ref_base = double(padarray(y_ref_base,[W/2 W/2],'replicate'));
    % Calculate the respective SNRs and PSNRs
    Base_snr(f-1) = 20*log10(std2(...
        y_current_base(N+1:base_Height+N,N+1:base_Width+N))/...
        std2(y_current_base(N+1:base_Height+N,N+1:base_Width+N)-Base));
    Enhanced_snr(f-1) = 20*log10(std2(y_current_enhance)/...
        std2(y_current_enhance-Enhanced));
    Base_psnr(f-1) = 20*log10(255/std2(...
        y_current_base(N+1:base_Height+N,N+1:base_Width+N)-Base));
    Enhanced_psnr(f-1) = 20*log10(255/std2(y_current_enhance - Enhanced));
end
%

figure,plot(F(2:end),Enhanced_snr,'kd',F(2:end),Base_snr,'ko','LineWidth',2)
xlim([F(2)-1 F(end)+1])
title('SNR (dB)')
legend('Enhanced Resolution','Base Resolution',0)
xlabel('Frame #'), ylabel('SNR (dB)')
%
figure,plot(F(2:end),Enhanced_psnr,'kd',F(2:end),Base_psnr,'ko','LineWidth',2)
xlim([F(2)-1 F(end)+1])
title('PSNR (dB)')
legend('Enhanced Resolution', 'Base Resolution', 0)
xlabel('Frame #'), ylabel('PSNR (dB)')