% QuantizationTable.m
% Design a Nonuniform Quantizer for an input signal
% with known pdf and generate optimum decision
% boundaries and output levels for a specified
% number of bits of quantization
%
% parameters to be specified:
% Width x Height = size of random image field to be generated
% B = number of bits of quantization
% PDFtype = type of pdf of the signal to be quantized
%   admissible types: 'uniform', 'gaussian' & 'laplacian'
% err = accuracy to terminate iteration
% The input signal is assumed to be zero-mean random variables(rv's)
% decision levels d(-k) = -d(k)
% similarly, output levels c(-k) = -c(k)

clear
close all
Height = 256;
Width = 256;
B = 4; % bits of quantization
PDFtype = 'laplacian'; % options: 'uniform', 'gaussian' & 'laplacian'
%
switch PDFtype
    case 'uniform'
        f = rand(Height,Width)-0.5;% zero-mean uniform pdf
        f = sqrt(12)*f(:);% var = 1
        f = f(logical(f>=0));
        fMax = max(f);
        err = 0.1e-02;
    case 'gaussian'
        f = randn(Height,Width);
        f = f(logical(f>=0));% only positive values of the r.v. are used
        f = f(:);
        fMax = max(f);
        err = 0.1e-03;
    case 'laplacian'
        f = rand(Height,Width);% uniform pdf
        f = (-1/sqrt(2))*log(1-f(:));% transformation for laplacian
        % r.v. with var = 1
        fMax = max(f);
        err = 0.1e-03;
end
%
Lf = length(f);
L = 2^B/2; % number of output levels
% choose an initial set of output levels
if strcmpi(PDFtype,'uniform')
    C = linspace(0,fMax,L);
else
    C = linspace(0,0.7*fMax,L);
end
PartitionRegion = zeros(L,Lf); % partitions consisting of input rv's
RegionCount = zeros(L,1); % number of elements in each partition
RegionSum = zeros(L,1); % sum of rv's in a given partition region
Distortion = zeros(L,1); % minimum distortion in a partition region
TotalDistortion = 1.0e+5; % total minimum distortion
% convergence factor is TD
TD = (TotalDistortion-sum(Distortion))/TotalDistortion;
% start iteration
It = 1; % iteration number
while TD >= err
    for r = 1:Lf
        d1 = abs(f(r)-C(1));
        J = 1;
        for k = 2:L % choose the nearest neighbor to C(k)
            d2 = abs(f(r)-C(k));
            if d2<d1
                d1 = d2;
                J = k;
            end
        end
        RegionCount(J) = RegionCount(J) + 1;
        RegionSum(J) = RegionSum(J) + f(r);
        PartitionRegion(J,r) = f(r);
        Distortion(J) = Distortion(J)+(f(r)-C(J))*(f(r)-C(J));
    end
    Index = logical(RegionCount == 0);% check for empty regions
    RegionCount(Index) = 1;% set empty regions to count 1
    Distortion = Distortion ./RegionCount;% average minimum distortions
    TD = (TotalDistortion-sum(Distortion))/TotalDistortion;
    C = RegionSum ./RegionCount; % refined output levels
    TotalDistortion = sum(Distortion);% total minimum distortion 
    It = It + 1;
end
clear Index
sprintf('Number of iterations = %d; \t bpp = %d',It,B)
% compute optimum decision boundaries from the optimum output levels
DR = zeros(L+1,1);
DR(L+1) = fMax;
for k = 2:L
    DR(k) = (C(k-1)+C(k))/2;
end
% plot output levels against decision regions
figure,stairs(DR(1:L),C,'k','LineWidth',2)
title(['Nonuniform quantizer: ' num2str(B) ' bpp' ' ' PDFtype ' pdf'])
xlabel('Decision Regions')
ylabel('Output Levels')
% decode using the codebook generated
f1 = zeros(size(f)); % quantized output
for r = 1:Lf
    d1 = abs(f(r)-C(1));
    for k = 2:L
        d2 = abs(f(r)-C(k));
        if d2<d1
            d1 = d2;
            J = k;
        end
    end
    f1(r) = C(J);
end 
% calculate mse between input and quantized output
mse = sum(Distortion);
sprintf('Number of levels = %d;\t MSE = %5.3f',L,mse)
% plot number of elements in each decision region as stem plot
figure,stem(RegionCount/(It-1),'k')
title(['Region Count at ' num2str(B) ' bpp' ' ' PDFtype ' pdf'])
xlabel('Level number'), ylabel('Count')
% print decision and output levels
sprintf('Decision boundaries \t Output levels \n')
sprintf('%5.3f\t\t %5.3f\n',DR(1:L),C)
