function ADPCM_Gain(A,dpcmType,B)
% ADPCM_Gain(A,dpcmType)
% Designs a 1st-order adaptive 1D DPCM (encoder only).
% The prediction error variance is adapted so that the 
% uniform quantizer is fixed. The adaptation procedure
% updates the current error pixel variance and accordingly
% adjusts the quantizer input (refer to Fig.6-11).


[Height,Width,Depth] = size(A);
if Depth > 1
    A1 = double(A(:,:,1));
else
    A1 = double(A);
end
%B = 3; % # bits of quantization
L = 2^B; % # levels of quantization
switch dpcmType
    case '1D'
        [alfa,E] = LinearPredict_1D(A1);% design the 1D 1st-order predictor
    case '2D'
        [alfa,E] = LinearPredict_2D(A1); % design the 2nd-order predictor
end
% Design a uniform quantizer with input as the prediction error
% with unit variance
E = E/std2(E);
dMin = mean2(E) - 3*std2(E);
dMax = mean2(E) + 3*std2(E);
q = 2*dMax/L;
q2 = q/2;
dR = linspace(dMin,dMax,L+1);
rL = zeros(L,1);
for k = 1:L
    rL(k) = dR(k)+q2;
end
% Adaptive rule: sigma(k)^2 = (1-gama)/(1-fB)*sigma(k-1)^2 +gama*e^2   
% Coefficients of adaptive algorithm
gama = 0.2; %b = 1.1711; fB = 2^(-b*B);
%a1 = (1-gama)/(1-fB);
a1 = (1-gama);
%
% Implement the DPCM
Mu = mean2(A1);% mean value of the image
pe = zeros(Height,Width);% array to store differential image
A2 = zeros(Height,Width);% array to store reconstructed image
peq = zeros(Height,Width);% array to store quantized differential image
A1 = A1 - Mu;% mean removed input image
SigVar = (std2(A1))^2;
switch dpcmType
    case '1D'
        for r = 1:Height
            A2(r,1) = A1(r,1) + Mu; % first pixel unquantized
            x1 = A1(r,1); % previous pixel
            varE = 0; % initial estimate of prediction error variance
            for c = 2:Width
                xhat = alfa*x1; % predicted pixel
                pe(r,c) = A1(r,c) - xhat; % differential pixel
                varE = a1*pe(r,c)*pe(r,c) + gama*varE;
                sigmaE = sqrt(varE);
                e1 = pe(r,c)/sigmaE;
                for k = 1:L
                    if e1>dR(k) && e1<=dR(k+1)
                        peq(r,c) = rL(k);
                    elseif e1<= dR(1)
                            peq(r,c) = rL(1);
                    elseif e1 > dR(L+1)
                        peq(r,c) = rL(L);
                    end
                end
                x1 = peq(r,c)*sigmaE + xhat; % previously reconstructed pixel
                A2(r,c) = x1 + Mu;% mean added reconstructed pixel
            end
        end
        SNR = 10*log10(SigVar/((std2(A1 - A2))^2));
    case '2D'
        x1 = zeros(Height+1,Width+2); % array to store reconstructed image
        A2(1,:) = A1(1,:) + Mu;
        A2(:,1) = A1(:,1) + Mu;
        A1 = padarray(A1,[1 2],'symmetric','pre');
        % First row, first column no prediction
        x1(1,:) = A1(1,:);% store previously reconstructed pixels
        x1(:,1) = A1(:,1);
        for r = 2:Height
            varE = 0;
            for c = 2:Width
                xhat = alfa(1)*x1(r,c-1) + alfa(2)*x1(r-1,c) + ...
                        alfa(3)*x1(r-1,c-1)+ alfa(4)*x1(r-1,c+1);
                pe(r,c) = A1(r,c) - xhat;
                varE = a1*pe(r,c)*pe(r,c) + gama*varE;
                sigmaE = sqrt(varE);
                e1 = pe(r,c)/sigmaE;
                for k = 1:L
                    if e1>dR(k) && e1<=dR(k+1)
                        peq(r,c) = rL(k);
                    elseif e1<= dR(1)
                            peq(r,c) = rL(1);
                    elseif e1 > dR(L+1)
                        peq(r,c) = rL(L);
                    end
                end
                x1(r,c) = peq(r,c)*sigmaE + xhat;% previously reconstructed pixel
                A2(r,c) = x1(r,c) + Mu;% mean added reconstructed pixel
            end
        end
        SNR = 10*log10(SigVar/(std2(A1(1:Height,1:Width) - A2)^2));
end
% display histogram of quantized & unquantized differential images
[H,B] = hist(pe(:),512);
H = H/sum(H);
[H1,B1] = hist(peq(:),512);
H1 = H1/sum(H1);
figure,subplot(2,1,1),plot(B,H,'k','LineWidth',2)
title(['Histogram of unquantized differential image: ' dpcmType])
ylabel('relative count')
subplot(2,1,2),plot(B1,H1,'k','LineWidth',2)
title(['Histogram of quantized differential image: ' dpcmType])
xlabel('pixel value'),ylabel('relative count')
% display the reconstructed image & calculate SNR
figure,imshow(pe,[]), title(['Unquantized differential image: ' dpcmType])
figure,imshow(peq,[]), title(['quantized differential image: ' dpcmType])
figure,imshow(A2,[]), title(['Reconstructed image: ' dpcmType])
sprintf('SNR = %4.2f\n',SNR)

    