function DPCM_2D(f,Qtype,B)
% DPCM_2D(f,Qtype,B)
% Designs and implements 2D DPCM where
% xhat = a1*x(m,n-1) + a2*x(m-1,n) + a3*x(m-1,n-1)...
% + a4*x(m-1,n+1)is the predictor
%
% Input:
%   f = intensity image to be encoded
%   Qtype = quantizer type: "uniform" or "nonuniform"
%   B = # bits of quantization
%
%   alfa = predictor coefficient
%   E = unquantized prediction error image
%   pe = unquantized differential image
%   peq = quantized differential image
%   y = DPCM reconstructed image
%
% This function implements only the encoder.


L = 2^B; % # levels in the quantizer
[Height,Width] = size(f);
%
% compute optimal predictor coefficients
[alfa,E] = LinearPredict_2D(f);
%
% Design the uniform quantizer using 5*std dev as the limits
switch Qtype
    case 'uniform'
        dMin = mean2(E) - 5*std2(E);
        dMax = mean2(E) + 5*std2(E);
        q = 2*dMax/L; % step size
        q2 = q/2;
        dR = linspace(dMin,dMax,L+1); % decision intervals
        rL = zeros(L,1); % reconstruction levels
        for k = 1:L
            rL(k) = dR(k)+q2;
        end
    case 'nonuniform'
        [DR,C] = dpcmQuantizer(E,B);% design a B-bit non-uniform quantizer
end
Mu = mean2(f);% mean value of the image
f = f - Mu;% remove mean value
% Implement the 2D DPCM
y = zeros(Height,Width);% array to store reconstructed image
pe = zeros(Height,Width);% array to store differential image
peq = zeros(Height,Width);% array to store quantized differential image
x1 = zeros(Height+1,Width+2); % array to store reconstructed image
y(1,:) = f(1,:) + Mu;
y(:,1) = f(:,1) + Mu;
%
f = padarray(f,[1 2],'symmetric','pre');
% First row, first column no prediction
x1(1,:) = f(1,:);% store previously reconstructed pixels
x1(:,1) = f(:,1);
for r = 2:Height
    for c = 2:Width
        xhat = alfa(1)*x1(r,c-1) + alfa(2)*x1(r-1,c) + ...
                alfa(3)*x1(r-1,c-1)+ alfa(4)*x1(r-1,c+1);
        pe(r,c) = f(r,c) - xhat;
        switch Qtype
            case 'uniform'
                for k = 1:L
                    if pe(r,c)>dR(k) && pe(r,c)<=dR(k+1)
                        peq(r,c) = rL(k);
                    elseif pe(r,c)<= dR(1)
                            peq(r,c) = rL(1);
                    elseif pe(r,c) > dR(L+1)
                        peq(r,c) = rL(L);
                    end
                end
            case 'nonuniform'
                %{
                for k = 1:L
                    if (pe(r,c)>DR(k) && pe(r,c)<=DR(k+1))
                            peq(r,c) = C(k);
                    end
                end
                %}
                d1 = abs(pe(r,c)-C(1));
                for k = 2:L
                    d2 = abs(pe(r,c)-C(k));
                    if d2<d1
                        d1 = d2;
                        J = k;
                    end
                end
                peq(r,c) = C(J); 
        end
        x1(r,c) = peq(r,c) + xhat;% previously reconstructed pixel
        y(r,c) = x1(r,c) + Mu;% mean added reconstructed pixel
    end
end
% Display differential and reconstructed images
figure,imshow(pe,[]), title('Differential image')
figure,imshow(y,[]), title('Reconstructed image')
[H,B] = hist(pe(:),512); % Histogram of differential image
H = H/sum(H);
[H1,B1] = hist(peq(:),512); % Histogram of quantized differential image
H1 = H1/sum(H1);
figure,subplot(2,1,1),plot(B,H,'k','LineWidth',2)
title('Histogram of unquantized differential image')
ylabel('relative count')
subplot(2,1,2),plot(B1,H1,'k','LineWidth',2)
title('Histogram of quantized differential image')
xlabel('pixel value'),ylabel('relative count')
% Compute SNR
SigVar = (std2(f(1:Height,1:Width)))^2;
SNR = 10*log10(SigVar/(std2(f(1:Height,1:Width) - y)^2));
sprintf('SNR = %4.2f\n',SNR)