function xq = deBlockingArt(A,Qscale)
% xq = deBlockingArt(A,Qscale)
% deblocking reduces blocking artifact due to JPEG compression
% Input:
%   A = Luma component of the input image, which will be compressed,
%       decompressed and smoothed.
%   Qscale = quantizer scale to be used in the DCT coefficient 
%            quantization.
% Output:
%   xq = smoothed image

N = 8;
[Height,Width] = size(A);
% Default JPEG luma quantization matrix
Qsteps = [16 11 10 16 24 40 51 61;...
    12 12 14 19 26 58 60 55;...
    14 13 16 24 40 57 69 56;...
    14 17 22 29 51 87 80 62;...
    18 22 37 56 68 109 103 77;...
    24 35 55 64 81 104 113 92;...
    49 64 78 87 103 121 120 101;...
    72 92 95 98 112 100 103 99];
%
Y = blkproc(double(A)-128,[N N],@dct2); % N x N 2D DCT of input image
Yq = zeros(Height,Width); % array to store quantized DCT coefficients
dctMean = zeros(N,N); % array for the mean of DCT coefficients
dctVar = zeros(N,N);  % array for the variance of DCT coefficients
dctAdjustType = 1; % 1 or 2
% Quantize, dequantize, adjust coefficients, and smooth blocks
switch dctAdjustType
    case 1
        % Quantize and calculate the mean of the absolute of the 
        % quantized DCT coefficients of the whole image
        for m = 1:N:Height
            for n = 1:N:Width
                temp = floor(Y(m:m+N-1,n:n+N-1)./(Qscale*Qsteps)...
                    + 0.5);
                dctMean = dctMean + abs(temp);
                Yq(m:m+N-1,n:n+N-1) = temp;
            end
        end
        dctMean = dctMean/((Height/N)*(Width/N));
        % Now adjust AC DCT coefficient amplitude only and
        % dequantize the DCT coefficients
        dctMean(1,1) = 0;
        mu = sum(dctMean(:))/63;
        for m = 1:N:Height
            for n = 1:N:Width
                temp = Yq(m:m+N-1,n:n+N-1);
                for k = 1:N
                    for l = 1:N
                        if k == 1 && l == 1
                        else
                            if temp(k,l)>0
                                temp(k,l) = temp(k,l) - 0.5 + mu-exp(-1/mu)*(1-exp(-1/mu));
                            elseif temp(k,l) < 0
                                temp(k,l) = temp(k,l) + 0.5 - mu+exp(-1/mu)*(1-exp(-1/mu));
                            end
                        end
                    end
                end
                Yq(m:m+N-1,n:n+N-1) = temp.*(Qscale*Qsteps);
            end
        end
        xq = blkproc(Yq,[N N],@idct2)+128; % 2D IDCT
        % perform smoothing operation
        dr = 0; dc = 0;
        for m = 1:N:Height
            for n = 1:N:Width
                temp = xq(m:m+N-1,n:n+N-1);
                dc = dc + sum(sum(A(m,n:n+N-1) - temp(1,:)));
                dc = dc + sum(sum(A(m+N-1,n:n+N-1) - temp(N,:)));
                dc = dc + sum(sum(A(m:m+N-1,n) - temp(:,1)));
                dc = dc + sum(sum(A(m:m+N-1,n+N-1) - temp(:,N)));
                dc = dc/(4*N);
                dr = dr + sum(sum((A(m,n:n+N-1) - temp(1,:)).^2));
                dr = dr + sum(sum((A(m+N-1,n:n+N-1) - temp(N,:)).^2));
                dr = dr + sum(sum((A(m:m+N-1,n) - temp(:,1)).^2));
                dr = dr + sum(sum((A(m:m+N-1,n+N-1) - temp(:,N)).^2));
                dr = dr/(4*N) - dc*dc;
                %T = 0.25/dr;% threshold for birds
                T = 1.25/dr; % threshold for Masuda
                if T < 1
                    alfa = 0.5*(1+sqrt(T));
                    if m == 1
                        if n == 1
                            i1 = xq(m+N-2,n:n+N-1);
                            i2 = xq(m+N,n:n+N-1);
                            xq(m+N-2,n:n+N-1)= alfa*i1 + (1-alfa)*i2;
                            xq(m+N,n:n+N-1)= alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n+N-2);
                            i2 = xq(m:m+N-1,n+N);
                            xq(m:m+N-1,n+N-2)= alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+N)= alfa*i2 + (1-alfa)*i1;
                        elseif n>1 && n<Width-N
                            i1 = xq(m+N-2,n:n+N-1);
                            i2 = xq(m+N,n:n+N-1);
                            xq(m+N-2,n:n+N-1)= alfa*i1 + (1-alfa)*i2;
                            xq(m+N,n:n+N-1)= alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n-1);
                            i2 = xq(m:m+N-1,n+1);
                            xq(m:m+N-1,n-1)= alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+1)= alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n+N-2);
                            i2 = xq(m:m+N-1,n+N);
                            xq(m:m+N-1,n+N-2) = alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+N) = alfa*i2 + (1-alfa)*i1;
                        else
                            i1 = xq(m:m+N-1,n-1);
                            i2 = xq(m:m+N-1,n+1);
                            xq(m:m+N-1,n-1)= alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+1)= alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m+N-2,n:n+N-1);
                            i2 = xq(m+N,n:n+N-1);
                            xq(m+N-2,n:n+N-1)= alfa*i1 + (1-alfa)*i2;
                            xq(m+N,n:n+N-1)= alfa*i2 + (1-alfa)*i1;
                        end
                    elseif m>1 && m<Height-N
                        if n == 1
                            i1 = xq(m-1,n:n+N-1);
                            i2 = xq(m+1,n:n+N-1);
                            xq(m-1,n:n+N-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m+1,n:n+N-1) = alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m+N-2,n:n+N-1);
                            i2 = xq(m+N,n:n+N-1);
                            xq(m+N-2,n:n+N-1)= alfa*i1 + (1-alfa)*i2;
                            xq(m+N,n:n+N-1)= alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n+N-2);
                            i2 = xq(m:m+N-1,n+N);
                            xq(m:m+N-1,n+N-2) = alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+N) = alfa*i2 + (1-alfa)*i1;
                        elseif n>1 && n<Width-N
                            i1 = xq(m-1,n:n+N-1);
                            i2 = xq(m+1,n:n+N-1);
                            xq(m-1,n:n+N-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m+1,n:n+N-1) = alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m+N-2,n:n+N-1);
                            i2 = xq(m+N,n:n+N-1);
                            xq(m+N-2,n:n+N-1)= alfa*i1 + (1-alfa)*i2;
                            xq(m+N,n:n+N-1)= alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n-1);
                            i2 = xq(m:m+N-1,n+1);
                            xq(m:m+N-1,n-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+1) = alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n+N-2);
                            i2 = xq(m:m+N-1,n+N);
                            xq(m:m+N-1,n+N-2) = alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+N) = alfa*i2 + (1-alfa)*i1;
                        else
                            i1 = xq(m-1,n:n+N-1);
                            i2 = xq(m+1,n:n+N-1);
                            xq(m-1,n:n+N-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m+1,n:n+N-1) = alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m+N-2,n:n+N-1);
                            i2 = xq(m+N,n:n+N-1);
                            xq(m+N-2,n:n+N-1)= alfa*i1 + (1-alfa)*i2;
                            xq(m+N,n:n+N-1)= alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n-1);
                            i2 = xq(m:m+N-1,n+1);
                            xq(m:m+N-1,n-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+1) = alfa*i2 + (1-alfa)*i1;
                        end
                    else
                        if n == 1
                            i1 = xq(m-1,n:n+N-1);
                            i2 = xq(m+1,n:n+N-1);
                            xq(m-1,n:n+N-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m+1,n:n+N-1) = alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n+N-2);
                            i2 = xq(m:m+N-1,n+N);
                            xq(m:m+N-1,n+N-2) = alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+N) = alfa*i2 + (1-alfa)*i1;
                        elseif n>1 && n<Width-N
                            i1 = xq(m-1,n:n+N-1);
                            i2 = xq(m+1,n:n+N-1);
                            xq(m-1,n:n+N-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m+1,n:n+N-1) = alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n-1);
                            i2 = xq(m:m+N-1,n+1);
                            xq(m:m+N-1,n-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+1) = alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n+N-2);
                            i2 = xq(m:m+N-1,n+N);
                            xq(m:m+N-1,n+N-2) = alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+N) = alfa*i2 + (1-alfa)*i1;
                        else
                            i1 = xq(m-1,n:n+N-1);
                            i2 = xq(m+1,n:n+N-1);
                            xq(m-1,n:n+N-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m+1,n:n+N-1) = alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n-1);
                            i2 = xq(m:m+N-1,n+1);
                            xq(m:m+N-1,n-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+1) = alfa*i2 + (1-alfa)*i1;
                        end
                    end
                end
                dr = 0; dc = 0;
            end
        end
    case 2
        % Alternative method of deblocking
        for m = 1:N:Height
            for n = 1:N:Width
                temp = floor(Y(m:m+N-1,n:n+N-1)./(Qscale*Qsteps)...
                    + 0.5);
                dctMean = dctMean + temp;
                dctVar = dctVar + temp.*temp;
                Yq(m:m+N-1,n:n+N-1) = temp;
            end
        end
        dctMean = dctMean/((Height/N)*(Width/N));
        dctVar = dctVar/((Height/N)*(Width/N)) - dctMean.^2;
        dctStd = sqrt(dctVar);
        dctStd(1,1) = 0;
        a = zeros(N,N);
        for k = 1:N
            for l = 1:N
                if dctStd(k,l) ~= 0
                    a(k,l) = sqrt(2)/dctStd(k,l);
                end
            end
        end
        [Indx,Indy] = find(a ~= 0);
        d = zeros(N,N);
        for i = 1:length(Indx)
            d(Indx(i),Indy(i)) = (Qsteps(Indx(i),Indy(i))/2)*...
                coth(a(Indx(i),Indy(i))*Qsteps(Indx(i),Indy(i))/2)...
                -1/a(Indx(i),Indy(i));
        end
        % Adjust DCT coefficient amplitude
        for m = 1:N:Height
            for n = 1:N:Width
                temp = Yq(m:m+N-1,n:n+N-1);
                Yq(m:m+N-1,n:n+N-1) = temp.*(Qscale*Qsteps)...
                    -sign(temp).*d;
            end
        end
        % Smooth the blocks
        xq = blkproc(Yq,[N N],@idct2)+128; % 2D IDCT
        dr = 0; dc = 0;
        for m = 1:N:Height
            for n = 1:N:Width
                temp = xq(m:m+N-1,n:n+N-1);
                dc = dc + sum(sum(A(m,n:n+N-1) - temp(1,:)));
                dc = dc + sum(sum(A(m+N-1,n:n+N-1) - temp(N,:)));
                dc = dc + sum(sum(A(m:m+N-1,n) - temp(:,1)));
                dc = dc + sum(sum(A(m:m+N-1,n+N-1) - temp(:,N)));
                dc = dc/(4*N);
                dr = dr + sum(sum((A(m,n:n+N-1) - temp(1,:)).^2));
                dr = dr + sum(sum((A(m+N-1,n:n+N-1) - temp(N,:)).^2));
                dr = dr + sum(sum((A(m:m+N-1,n) - temp(:,1)).^2));
                dr = dr + sum(sum((A(m:m+N-1,n+N-1) - temp(:,N)).^2));
                dr = dr/(4*N) - dc*dc;
                T = 1.25/dr;
                %T = 0.25/dr;
                if T < 1
                    alfa = 0.5*(1+sqrt(T));
                    if m == 1
                        if n == 1
                            i1 = xq(m+N-2,n:n+N-1);
                            i2 = xq(m+N,n:n+N-1);
                            xq(m+N-2,n:n+N-1)= alfa*i1 + (1-alfa)*i2;
                            xq(m+N,n:n+N-1)= alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n+N-2);
                            i2 = xq(m:m+N-1,n+N);
                            xq(m:m+N-1,n+N-2)= alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+N)= alfa*i2 + (1-alfa)*i1;
                        elseif n>1 && n<Width-N
                            i1 = xq(m+N-2,n:n+N-1);
                            i2 = xq(m+N,n:n+N-1);
                            xq(m+N-2,n:n+N-1)= alfa*i1 + (1-alfa)*i2;
                            xq(m+N,n:n+N-1)= alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n-1);
                            i2 = xq(m:m+N-1,n+1);
                            xq(m:m+N-1,n-1)= alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+1)= alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n+N-2);
                            i2 = xq(m:m+N-1,n+N);
                            xq(m:m+N-1,n+N-2) = alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+N) = alfa*i2 + (1-alfa)*i1;
                        else
                            i1 = xq(m:m+N-1,n-1);
                            i2 = xq(m:m+N-1,n+1);
                            xq(m:m+N-1,n-1)= alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+1)= alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m+N-2,n:n+N-1);
                            i2 = xq(m+N,n:n+N-1);
                            xq(m+N-2,n:n+N-1)= alfa*i1 + (1-alfa)*i2;
                            xq(m+N,n:n+N-1)= alfa*i2 + (1-alfa)*i1;
                        end
                    elseif m>1 && m<Height-N
                        if n == 1
                            i1 = xq(m-1,n:n+N-1);
                            i2 = xq(m+1,n:n+N-1);
                            xq(m-1,n:n+N-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m+1,n:n+N-1) = alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m+N-2,n:n+N-1);
                            i2 = xq(m+N,n:n+N-1);
                            xq(m+N-2,n:n+N-1)= alfa*i1 + (1-alfa)*i2;
                            xq(m+N,n:n+N-1)= alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n+N-2);
                            i2 = xq(m:m+N-1,n+N);
                            xq(m:m+N-1,n+N-2) = alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+N) = alfa*i2 + (1-alfa)*i1;
                        elseif n>1 && n<Width-N
                            i1 = xq(m-1,n:n+N-1);
                            i2 = xq(m+1,n:n+N-1);
                            xq(m-1,n:n+N-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m+1,n:n+N-1) = alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m+N-2,n:n+N-1);
                            i2 = xq(m+N,n:n+N-1);
                            xq(m+N-2,n:n+N-1)= alfa*i1 + (1-alfa)*i2;
                            xq(m+N,n:n+N-1)= alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n-1);
                            i2 = xq(m:m+N-1,n+1);
                            xq(m:m+N-1,n-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+1) = alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n+N-2);
                            i2 = xq(m:m+N-1,n+N);
                            xq(m:m+N-1,n+N-2) = alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+N) = alfa*i2 + (1-alfa)*i1;
                        else
                            i1 = xq(m-1,n:n+N-1);
                            i2 = xq(m+1,n:n+N-1);
                            xq(m-1,n:n+N-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m+1,n:n+N-1) = alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m+N-2,n:n+N-1);
                            i2 = xq(m+N,n:n+N-1);
                            xq(m+N-2,n:n+N-1)= alfa*i1 + (1-alfa)*i2;
                            xq(m+N,n:n+N-1)= alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n-1);
                            i2 = xq(m:m+N-1,n+1);
                            xq(m:m+N-1,n-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+1) = alfa*i2 + (1-alfa)*i1;
                        end
                    else
                        if n == 1
                            i1 = xq(m-1,n:n+N-1);
                            i2 = xq(m+1,n:n+N-1);
                            xq(m-1,n:n+N-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m+1,n:n+N-1) = alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n+N-2);
                            i2 = xq(m:m+N-1,n+N);
                            xq(m:m+N-1,n+N-2) = alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+N) = alfa*i2 + (1-alfa)*i1;
                        elseif n>1 && n<Width-N
                            i1 = xq(m-1,n:n+N-1);
                            i2 = xq(m+1,n:n+N-1);
                            xq(m-1,n:n+N-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m+1,n:n+N-1) = alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n-1);
                            i2 = xq(m:m+N-1,n+1);
                            xq(m:m+N-1,n-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+1) = alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n+N-2);
                            i2 = xq(m:m+N-1,n+N);
                            xq(m:m+N-1,n+N-2) = alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+N) = alfa*i2 + (1-alfa)*i1;
                        else
                            i1 = xq(m-1,n:n+N-1);
                            i2 = xq(m+1,n:n+N-1);
                            xq(m-1,n:n+N-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m+1,n:n+N-1) = alfa*i2 + (1-alfa)*i1;
                            i1 = xq(m:m+N-1,n-1);
                            i2 = xq(m:m+N-1,n+1);
                            xq(m:m+N-1,n-1) = alfa*i1 + (1-alfa)*i2;
                            xq(m:m+N-1,n+1) = alfa*i2 + (1-alfa)*i1;
                        end
                    end
                end
                dr = 0; dc = 0;
            end
        end
end           