% Example8_1.m
% Image compression in the wavelet domain
% An L-level 2D DWT of an input image is computed
% using "db8" orthogonal wavelet. Optimal quantizer
% bits are computed for a bit budget of R (typically 1) bpp
% and then the quantization steps are computed.
% Uniform quantization is used across all the bands.
% if the input image RGB, it is converted to YCbCr
% components, Cb ans Cr components are down sampled to
% 4:2:0 or 4:2:2 format and quantized. 
% User may specify the values for R, L, and sampling format.
% Parameters are:
%   SamplingFormat = "4:2:0" or "4:2:2"
%   L = number of levels of DWT
%   R = bit budget in bpp
%   BitAssignRule = "optimal" or "integer"

clear
%A = imread('masuda2.ras');
A = imread('birds.ras');
L = 3; % number of DWT levels
% make sure that the image size is divisible by NL
[x,y,z] = size(A);
if mod(x,2^L) ~=0
    Height = floor(x/(2^L))*(2^L);
else
    Height = x;
end
if mod(y,2^L) ~=0
    Width = floor(y/(2^L))*(2^L);
else
    Width = y;
end
Depth = z;
clear x y z
%
SamplingFormat = '4:2:2'; % for RGB image
wName = 'db2'; % wavelet name. See MATLAB for available wavelets
%wName = 'bior1.5';
%wName = 'coif2'; 
% bit budget: R1 for Y, R2 for Cb & R3 for Cr
R1 = 1.0; R2 = 0.5; R3 = 0.5;
BitAssignRule = 'optimal';
if Depth == 1
    % input image is B/W
    Y = double(A(1:Height,1:Width))-128;
else
    % Input image is RGB. Convert it to YCbCr
    % using JPEG2000 reversible component transformation
    A1 = double(A(1:Height,1:Width,1:Depth))-128;
    Y = 0.299*A1(:,:,1) + 0.587*A1(:,:,2) + 0.144*A1(:,:,3);
    Cb = -0.16875*A1(:,:,1) - 0.33126*A1(:,:,2) + 0.5*A1(:,:,3);
    Cr = 0.5*A1(:,:,1) - 0.41869*A1(:,:,2) - 0.08131*A1(:,:,3);
    switch SamplingFormat
        % subsample chroma components
        case '4:2:0'
            Cb = imresize(Cb,[Height/2 Width/2],'cubic');
            Cr = imresize(Cr,[Height/2 Width/2],'cubic');
        case '4:2:2'
            Cb = imresize(Cb,[Height Width/2],'cubic');
            Cr = imresize(Cr,[Height Width/2],'cubic');
    end
end
%
[C,S] = wavedec2(Y,L,wName); % L-level 2D DWT of the Luma
% Compute optimal quantizer bits
switch BitAssignRule
    case 'optimal'
        [Qbits,Qsteps] = DWToptimalBits(C,S,wName,R1);
    case 'integer'
        [Qbits,Qsteps] = AssignIntgrBits2DWT(C,S,wName,R1);
end
% find total bits for the Y component
TotalBitsY = (S(L+1,1)*S(L+1,2))*sum(Qbits(1:3))+...
    (S(L,1)*S(L,2))*sum(Qbits(4:6))+(S(L-1,1)*S(L-1,2))*sum(Qbits(7:10));
sprintf('Number of levels of DWT = %d\nquantizer %s bits\n',L,BitAssignRule)
disp(Qbits(1:3*L+1)')
%
% Quantize and dequantize the coefficients
% Function "quantizeDWT" quantizes and dequantizes the DWT
% coefficients
Cq = quantizeDWT(C,S,Qsteps);
Yq = waverec2(Cq,S,wName); % do inverse 2D DWT of quantized Y coefficients
SNR = 20*log10(std2(Y)/std2(Y-Yq));
% Calculate SNR of Y component
sprintf('SNR(Y) = %4.2fdB\n',SNR)
if Depth == 1
    figure,imshow(Yq+128,[])
    AvgBit = TotalBitsY/(Height*Width);
    sprintf('Desired avg. rate = %3.2f bpp\tActual avg. rate = %3.2f bpp\n',R1,AvgBit)
end
% If the input image is RGB, quantize the Cb & Cr components
if Depth > 1
    [C,S] = wavedec2(Cb,L,wName); % L-level 2D DWT of Cb
    switch BitAssignRule
        case 'optimal'
            [Qbits,Qsteps] = DWToptimalBits(C,S,wName,R2);
        case 'integer'
            [Qbits,Qsteps] = AssignIntgrBits2DWT(C,S,wName,R2);
    end
    % find total bits for the Cb component
    TotalBitsCb = (S(L+1,1)*S(L+1,2))*sum(Qbits(1:3))+...
    (S(L,1)*S(L,2))*sum(Qbits(4:6))+(S(L-1,1)*S(L-1,2))*sum(Qbits(7:10));
    sprintf('Cb quantizer bits: ')
    disp(Qbits(1:3*L+1)')
    % quantize Cb DWT coefficients
    Cq = quantizeDWT(C,S,Qsteps);
    Cbq = waverec2(Cq,S,wName);% do 2D IDWT of quantized Cb
    SNRcb = 20*log10(std2(Cb)/std2(Cb-Cbq));
    %
    [C,S] = wavedec2(Cr,L,wName); % L-level 2D DWT of Cr
    switch BitAssignRule
        case 'optimal'
            [Qbits,Qsteps] = DWToptimalBits(C,S,wName,R3);
        case 'integer'
            [Qbits,Qsteps] = AssignIntgrBits2DWT(C,S,wName,R3);
    end
    % find total bits for the Cr component
    TotalBitsCr = (S(L+1,1)*S(L+1,2))*sum(Qbits(1:3))+...
    (S(L,1)*S(L,2))*sum(Qbits(4:6))+(S(L-1,1)*S(L-1,2))*sum(Qbits(7:10));
    sprintf('\nCr quantizer bits: ')
    disp(Qbits(1:3*L+1)')
    % Find the overall average bit rate in bpp.
    AvgBit = (TotalBitsY+TotalBitsCb+TotalBitsCr)/(Height*Width);
    sprintf('Actual avg. rate = %5.4f bpp\n',AvgBit)
    % quantize Cr DWT coefficients
    Cq = quantizeDWT(C,S,Qsteps); 
    Crq = waverec2(Cq,S,wName); % do 2D IDWT of quantized Cr coefficients
    SNRcr = 20*log10(std2(Cr)/std2(Cr-Crq));
    % upsample Cb & Cr to full resolution
    Cbq = imresize(Cbq,[Height Width],'cubic');
    Crq = imresize(Crq,[Height Width],'cubic');
    Ahat = zeros(Height,Width,Depth);
    Ahat(:,:,1) = Yq + 1.402*Crq + 128;
    Ahat(:,:,2) = Yq - 0.34413*Cbq - 0.71414*Crq + 128;
    Ahat(:,:,3) = Yq + 1.772*Cbq + 128;
    figure,imshow(uint8(round(Ahat)))
    sprintf('Chroma sampling = %s\n',SamplingFormat)
    sprintf('SNR(Cb) = %4.2fdB\tSNR(Cr) = %4.2fdB\n',SNRcb,SNRcr)
end    