% JPG2K_Tiles.m
% Implements JPEG2000 wavelet transform-based compression.
% Implements only the scalar quantization normative of the
% standard. Implements up to NL levels of 2D DWT using
% Daubechies' 9/7 filter bank.
% Implements 4:2:0 and 4:2:2 sampling formats specified
% by the user.
%
clear
NL = 3; % number of levels of 2D DWT
SamplingFormat = '4:2:0';
%A = imread('cameraman.tif');
A = imread('masuda2.ras');
% make sure that the image size is divisible by NL
[X,Y,Z] = size(A);
if mod(X,2^NL) ~=0
    Height = floor(X/(2^NL))*(2^NL);
else
    Height = X;
end
if mod(Y,2^NL) ~=0
    Width = floor(Y/(2^NL))*(2^NL);
else
    Width = Y;
end
Depth = Z;
clear X Y Z
TileSizeV = Height/2;
TileSizeH = Width/2;
% Do forward Irreversible Component Transformation (ICT)
% In JPEG2000 jargon, y0 corresponds to the Luma Y,
% y1 to Cb and y2 to Cr.
if Depth == 1
    y0 = double(A(1:Height,1:Width)) - 128; % DC level shift
else
    A1 = double(A(1:Height,1:Width,:))-128; % DC level shift
    y0 = 0.299*A1(:,:,1) + 0.587*A1(:,:,2) + 0.144*A1(:,:,3);
    y1 = -0.16875*A1(:,:,1) - 0.33126*A1(:,:,2) + 0.5*A1(:,:,3);
    y2 = 0.5*A1(:,:,1) - 0.41869*A1(:,:,2) - 0.08131*A1(:,:,3);
    clear A1
    switch SamplingFormat
        case '4:2:0'
            y1 = imresize(y1,[Height/2 Width/2],'cubic');
            y2 = imresize(y2,[Height/2 Width/2],'cubic');
        case '4:2:2'
            y1 = imresize(y1,[Height Width/2],'cubic');
            y2 = imresize(y2,[Height Width/2],'cubic');
    end
end
% Daubechies 9/7 Filters
% Note: The Daubechies 9/7 filterbank has 9 taps for its analysis
% Lowpass and synthesis highpass FIR filters;
% 7 taps for its analysis highpass and synthesis lowpass
% FIR filters. However, in order to obtain subbands that are
% exact multiples of 2^-n, n being a positive integer, the
% number of taps in the filters must be even. That is why you
% will notice the filter lengths to be 10 instead of 9 or 7.
% The extra tap values are zero.
%
% LO_D = analysis lowpass filter
% LO_R = synthesis lowpass filter
% HI_D = analysis highpass filter
% HI_R = synthesis highpass filter
LO_D = [0 0.0267487574 -0.0168641184 -0.0782232665 0.2668641184...
    0.6029490182 0.2668641184 -0.0782232665 -0.0168641184...
    0.0267487574];
HI_D = [0 0.0912717631 -0.0575435263 -0.5912717631...
    1.1150870525 -0.5912717631 -0.0575435263 0.0912717631 0 0];
%
LO_R = [0 -0.0912717631 -0.0575435263 0.5912717631...
    1.1150870525 0.5912717631 -0.0575435263 -0.0912717631 0 0];
HI_R = [0 0.0267487574 0.0168641184 -0.0782232665 -0.2668641184...
    0.6029490182 -0.2668641184 -0.0782232665 0.0168641184...
    0.0267487574];
%
% "dwt_Coef" cell array has NL x 4 cells.
% In each cell at level NL, the 1st component is the LL coefficients,
% 2nd component is HL coefficients, 3rd component LH and
% 4th component HH coefficients.
%
% Do an NL-level 2D DWT of the Y component
dwt_y0 = cell(NL,4);
idwt_y0 = cell(NL,1);
Y0 = zeros(Height,Width);
TotalBits = 0;
for r = 1:TileSizeV:Height
    for c = 1:TileSizeH:Width
        for n = 1:NL
            if n == 1
                [dwt_y0{n,1},dwt_y0{n,2},dwt_y0{n,3},dwt_y0{n,4}] =...
                    dwt2(y0(r:r+TileSizeV-1,c:c+TileSizeH-1),LO_D,HI_D,'mode','per');
            else
                [dwt_y0{n,1},dwt_y0{n,2},dwt_y0{n,3},dwt_y0{n,4}] =...
                    dwt2(dwt_y0{n-1,1},LO_D,HI_D,'mode','per');
            end
        end
        % Quantize the Y DWT coefficients
        [dwt_y0,y0Bits] = jpg2K_Quantize(dwt_y0,NL,1);
        TotalBits = TotalBits + y0Bits;
        %
        % Do an NL-level 2D IDWT of the Y component
        % The cell array idwt_img has NL cells with each cell corresponding to
        % the reconstructed LL image at that level.
        for n = NL:-1:1
            if n == 3
                idwt_y0{n} = idwt2(dwt_y0{n,1},dwt_y0{n,2},dwt_y0{n,3},...
                    dwt_y0{n,4},LO_R,HI_R,'mode','per');
            else
                idwt_y0{n} = idwt2(idwt_y0{n+1},dwt_y0{n,2},dwt_y0{n,3},...
                    dwt_y0{n,4},LO_R,HI_R,'mode','per');
            end
        end
        Y0(r:r+TileSizeV-1,c:c+TileSizeH-1)=idwt_y0{1};
    end
end
% Interpolate to remove tile boundaries
%Y0(TileSizeV-1,1:Width) = Y0(TileSizeV-2,1:Width);
%Y0(TileSizeV+1,1:Width) = Y0(TileSizeV+2,1:Width);
Y0(TileSizeV,1:Width) = (Y0(TileSizeV-2,1:Width)+Y0(TileSizeV+2,1:Width))/2;
%Y0(:,TileSizeH-1) = Y0(:,TileSizeH-5);
%Y0(:,TileSizeH+1) = Y0(:,TileSizeH+5);
Y0(:,TileSizeH)= (Y0(:,TileSizeH-2)+Y0(:,TileSizeH+2))/2;
SNRy0 = 20*log10(std2(y0)/std2(y0-Y0));
if Depth > 1
    switch SamplingFormat
        case '4:2:0'
            H = Height/2;
            W = Width/2;
            TV = TileSizeV/2;
            TH = TileSizeH/2;
        case '4:2:2'
            H = Height;
            W = Width/2;
            TV = TileSizeV;
            TH = TileSizeH/2;
    end
    % Do an NL-level 2D DWT of the Cb and Cr components
    dwt_y1 = cell(NL,4);
    dwt_y2 = cell(NL,4);
    for r = 1:TV:H
        for c = 1:TH:W
            for n = 1:NL
                if n == 1
                    [dwt_y1{n,1},dwt_y1{n,2},dwt_y1{n,3},dwt_y1{n,4}] =...
                        dwt2(y1(r:r+TV-1,c:c+TH-1),LO_D,HI_D,'mode','per');
                    [dwt_y2{n,1},dwt_y2{n,2},dwt_y2{n,3},dwt_y2{n,4}] =...
                        dwt2(y2(r:r+TV-1,c:c+TH-1),LO_D,HI_D,'mode','per');
                else
                    [dwt_y1{n,1},dwt_y1{n,2},dwt_y1{n,3},dwt_y1{n,4}] =...
                        dwt2(dwt_y1{n-1,1},LO_D,HI_D,'mode','per');
                    [dwt_y2{n,1},dwt_y2{n,2},dwt_y2{n,3},dwt_y2{n,4}] =...
                        dwt2(dwt_y2{n-1,1},LO_D,HI_D,'mode','per');
                end
            end
            %
            % Quantize the Cb and Cr DWT coefficients
            [dwt_y1,y1Bits] = jpg2K_Quantize(dwt_y1,NL,2);
            [dwt_y2,y2Bits] = jpg2K_Quantize(dwt_y2,NL,3);
            TotalBits = TotalBits + y1Bits + y2Bits;
            %
            % % Do an NL-level 2D IDWT of the Cb and Cr components
            idwt_y1 = cell(NL,1);
            idwt_y2 = cell(NL,1);
            %
            for n = NL:-1:1
                if n == 3
                    idwt_y1{n} = idwt2(dwt_y1{n,1},dwt_y1{n,2},dwt_y1{n,3},...
                        dwt_y1{n,4},LO_R,HI_R,'mode','per');
                    idwt_y2{n} = idwt2(dwt_y2{n,1},dwt_y2{n,2},dwt_y2{n,3},...
                        dwt_y2{n,4},LO_R,HI_R,'mode','per');
                else
                    idwt_y1{n} = idwt2(idwt_y1{n+1},dwt_y1{n,2},dwt_y1{n,3},...
                        dwt_y1{n,4},LO_R,HI_R,'mode','per');
                    idwt_y2{n} = idwt2(idwt_y2{n+1},dwt_y2{n,2},dwt_y2{n,3},...
                        dwt_y2{n,4},LO_R,HI_R,'mode','per');
                end
            end
            Y1(r:r+TV-1,c:c+TH-1)=idwt_y1{1};
            Y2(r:r+TV-1,c:c+TH-1)=idwt_y2{1};
        end
    end
    z = [Y1(TV-1,:);Y1(TV+1,:)];
    Y1(TV,:) = imresize(z,[1 size(Y1,2)],'bicubic');
    z = [Y2(TV-1,:);Y2(TV+1,:)];
    Y2(TV,:) = imresize(z,[1 size(Y1,2)],'bicubic');
    z = [Y1(:,TH-1) Y1(:,TH+1)];
    Y1(:,TH) = imresize(z,[size(Y1,1) 1],'bicubic');
    z = [Y2(:,TH-1) Y2(:,TH+1)];
    Y2(:,TH) = imresize(z,[size(Y2,1) 1],'bicubic');
    % Compute the SNR of Cb and Cr components due to quantization
    SNRy1 = 20*log10(std2(y1)/std2(y1-Y1));
    SNRy2 = 20*log10(std2(y2)/std2(y2-Y2));
    % Upsample Cb & Cr components to original full size
    Y1 = imresize(Y1,[Height Width],'cubic');
    Y2 = imresize(Y2,[Height Width],'cubic');
end
% Do inverse ICT & level shift
if Depth == 1
    Y0 = Y0 + 128;
    figure,imshow(uint8(Y0))
    sprintf('Average bit rate = %5.2f bpp\n',TotalBits/(Height*Width))
    sprintf('SNR(Y) = %4.2f dB\n',SNRy0)
else
    Ahat = zeros(Height,Width,Depth);
    Ahat(:,:,1) = Y0 + 1.402*Y2 + 128;
    Ahat(:,:,2) = Y0 - 0.34413*Y1 - 0.71414*Y2 + 128;
    Ahat(:,:,3) = Y0 + 1.772*Y1 + 128;
    figure,imshow(uint8(Ahat))
    sprintf('Average bit rate = %5.2f bpp\n',...
        TotalBits/(Height*Width))
    sprintf('SNR(Y) = %4.2fdB\tSNR(Cb) = %4.2f dB\tSNR(Cr) = %4.2fdB\n',...
        SNRy0,SNRy1,SNRy2)
end