function [x,y,pf] = MCpyramid_Full(A,B)
% [x,y,pf] = M Cpyramid_Full(A,B)
% Computes motion vectors of 8 x 8 moving blocks 
% in an intensity image using hierarchical search and
% does a motion compensated prediction to a single pel accuracy
% Input:
%   A = reference frame - cell array w/ 3 cells
%   B = current frame - cell array w/ 3 cells
%
% Outpur:
%   x = horizontal component of motion vector
%   y = Vertical component of motion vector
%   pf = motion compensated prediction image, same size as input image


% Input images A & B each have 3 levels in their pyramids
% Level 1 is the full resolution image of size Height x Width
% Level 2 is the next lower resolution of size Height/2 x Width/2
% Level 3 is the lowest resolution image of size Height/4 x Width/4
%
% pad input images on left, right, top, and bottom
A1{3} = double(padarray(A{3},[8 8],'symmetric'));
B1{3} = double(padarray(B{3},[8 8],'symmetric'));
A1{2} = double(padarray(A{2},[8 8],'symmetric'));
B1{2} = double(padarray(B{2},[8 8],'symmetric'));
A1{1} = double(padarray(A{1},[8 8],'symmetric'));
B1{1} = double(padarray(B{1},[8 8],'symmetric'));
% Start with block size 2 x 2
N = 2; L = 3;
% x & y are cell arrays w/ 3 elements to store motion vector components
% corresponding to the three levels
x = cell(L,1); y = cell(L,1);
[Height,Width] = size(A{3});% Initial image size corresponding to level 3
x{L} = int16(zeros(Height/N,Width/N));
y{L} = int16(zeros(Height/N,Width/N));
%
% Estimate motion at level 3
for r = N+6:N:Height
    rblk = floor(r/N);
    for c = N+6:N:Width
        cblk = floor(c/N);
        D = 1.0e+10;% initial city block distance
        for u = -2:2
            for v = -2:2
                Indr1 = r+u+1;
                Indr2 = Indr1 + 1;
                Indc1 = c+v+1;
                Indc2 = Indc1 + 1;
                d = B1{L}(r+1:r+N,c+1:c+N)-A1{L}(Indr1:Indr2,Indc1:Indc2);
                d = sum(abs(d(:)));% city block distance between pixels
                if d < D
                    D = d;
                    x{L}(rblk,cblk) = v; y{L}(rblk,cblk) = u;
                end
            end
        end
    end
end
% Estimate motion at level 2
% block size is 4 x 4
N = 4; L = 2;
Height = 2*Height; Width = 2*Width; % double the height & width
for r = N+4:N:Height
    rblk = floor(r/N);
    for c = N+4:N:Width
        cblk = floor(c/N);
        D = 1.0e+10;% initial city block distance
        x1 = x{L+1}(rblk,cblk)/4; y1 = y{L+1}(rblk,cblk)/4;
        for u = -2:2
            for v = -2:2
                Indr1 = r+u+y1+1;
                Indr2 = Indr1 + 3;
                Indc1 = c+v+x1+1;
                Indc2 = Indc1 + 3;
                d = B1{L}(r+1:r+N,c+1:c+N)-A1{L}(Indr1:Indr2,Indc1:Indc2);
                d = sum(abs(d(:)));% city block distance between pixels
                if d < D
                    D = d;
                    x{L}(rblk,cblk) = v; y{L}(rblk,cblk) = u;
                end
            end
        end
    end
end
% Estimate motion at level 1 (highest resolution)
% block size is 8 x 8
N = 8; L = 1;
Height = 2*Height; Width = 2*Width; % double the height & width
for r = N:N:Height
    rblk = floor(r/N);
    for c = N:N:Width
        cblk = floor(c/N);
        D = 1.0e+10;% initial city block distance
        x1 = x{L+1}(rblk,cblk)/4; y1 = y{L+1}(rblk,cblk)/4;
        for u = -6:6
            for v = -6:6
                Indr1 = round(r+u+y1+1);
                Indr2 = Indr1 + 7;
                Indc1 = round(c+v+x1+1);
                Indc2 = Indc1 + 7;
                d = B1{L}(r+1:r+N,c+1:c+N)-A1{L}(Indr1:Indr2,Indc1:Indc2);
                d = sum(abs(d(:)));% city block distance between pixels
                if d < D
                    D = d;
                    x{L}(rblk,cblk) = v; y{L}(rblk,cblk) = u;
                end
            end
        end
    end
end
% Reconstruct current frame using prediction error & reference frame
N = 8;
[Height,Width] = size(A{1});
N2 = 2*N;
pf = double(zeros(Height,Width)); % prediction frame
Br = double(zeros(Height,Width)); % reconstructed frame
for r = 1:N:Height
    rblk = floor(r/N) + 1;
    for c = 1:N:Width
        cblk = floor(c/N) + 1;
        x1 = x{1}(rblk,cblk); y1 = y{1}(rblk,cblk);
        pf(r:r+N-1,c:c+N-1) = B1{1}(r+N:r+N2-1,c+N:c+N2-1)...
            -A1{1}(r+N+y1:r+y1+N2-1,c+N+x1:c+x1+N2-1);
        Br(r:r+N-1,c:c+N-1) = A1{1}(r+N+y1:r+y1+N2-1,c+N+x1:c+x1+N2-1)...
            + pf(r:r+N-1,c:c+N-1);
    end
end
%
figure,imshow(uint8(round(Br))),title('Reconstructed image')